﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ContinuationExamples
{
    class Program
    {
        static void Main(string[] args)
        {
            ///A way to call async methods in Main
            Task.Run(async () =>
            {
                Console.WriteLine("In Main before await");
                await MainAsync(args); //należy stworzyć sobie asynchronicznego Maina/you need to create an asynchronous Main
                Console.WriteLine("In Main after await");
            }).GetAwaiter().GetResult();//and wait for it to finish/i poczekać na jego zakończenie
            Console.WriteLine("We can finish here / Tu możemy już kończyć");
            Console.ReadLine();
        }
        public static async Task MainAsync(string[] args) //when I have at least one await call, the method can be asynchronous
        //public static void Main(string[] args) //when we don't have await we have to do the old way
        {
            #region The simplest continuation example
            Console.WriteLine("\n--------------------Simple continuation-------------------");
            Task firstTask = Task.Factory.StartNew(() =>
            {
                Console.WriteLine("First task has started");
                Thread.Sleep(1000);
                Console.WriteLine("First task has finished");
            }
            );
            Task secondTask = firstTask.ContinueWith(ant =>
            {
                Console.WriteLine("Second task has started");
                Thread.Sleep(1000);
                Console.WriteLine("Second task has finished");
            }
            );
            await secondTask;
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion




            #region Continuation with the passing values
            Console.WriteLine("\n--------------------Continuation with the passing values--");
            Task<int> task1a = Task.Factory.StartNew<int>(() => { Thread.Sleep(1000); return 1; }); 
            Task<int> task2a = task1a.ContinueWith<int>(ant => ant.Result + 1);
            Console.WriteLine($"Result of task1a: {task1a.Result}");
            Console.WriteLine($"Result of task2a: {task2a.Result}");
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion

            #region ContinueWith
            Console.WriteLine("\n --------------One more example of ContinueWith-------");
            //intTask = Task.Run(() =>
            //{
            //    Thread.Sleep(1000);
            //    return 5;
            //}).ContinueWith((x) => x.Result + 5);

            //you can also this way
            Task<int> intTask = Task.Run(() =>
            {
                Thread.Sleep(1000);
                return 12;
            });
            var continueTask = intTask.ContinueWith((x) => { Thread.Sleep(1000); return x.Result / 2; }); //just added sleep here to show that the second task is also doing something
            Console.WriteLine($"We have got result = {intTask.Result}");
            Console.WriteLine($"We have got result2 = {continueTask.Result}");
            #endregion

            #region Continuation with the passing values v2
            Console.WriteLine("\n--------------------Continuation with the passing values v2--");
            int res = Task.Factory.StartNew<int>(() => 1)
                    .ContinueWith<int>(ant => ant.Result + 1)
                    .Result;
            Console.WriteLine($"Result of the second version: {res}");
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion


            #region Continuation with the exception
            Console.WriteLine("\n--------------------Continuation with the exception----------");
            Task<int> task1ex = Task.Factory.StartNew<int>(() => { throw new Exception("The first one threw an exception"); });
            //Task<int> task2ex = task1ex.ContinueWith<int>(ant => {Console.WriteLine("Exception: " + ant.Exception.Message); return 0; });
            //A safe patern states that we should forward an exception to the place where we expect the result
            Task<int> task2ex = task1ex.ContinueWith<int>(ant => { if (ant.Exception != null) throw ant.Exception; return 0; });
            try
            {
                Console.WriteLine($"Result of task2ex: {task2ex.Result}");
            }
            catch (AggregateException ae)  
            {
                Console.WriteLine("AggregateException has message: " + ae.Message);
                foreach (var ex in ae.InnerExceptions)
                {
                    Console.WriteLine("I caught exception: " + ex.Message);
                    Console.WriteLine("Inner exception:" + ex.InnerException.Message);
                }
            }
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion


            #region Different paths for exceptions
            Console.WriteLine("\n--------------------Different paths for exceptions----------");
            Task<int> task1r = Task.Factory.StartNew<int>(() => {
                //uncomment to simulate error
                //throw new Exception("The first trew an exception");
                return 1;
            });
            Task<int> taskErr = task1r.ContinueWith<int>(ant => { Console.WriteLine("Exception: " + ant.Exception.Message); return 1; },TaskContinuationOptions.OnlyOnFaulted);
            Task<int> taskNotErr = task1r.ContinueWith<int>(ant => { Console.WriteLine("There was no error") ; return ant.Result+2; }, TaskContinuationOptions.NotOnFaulted);
            try
            {
                if (taskErr != null) await taskErr;
                if (taskNotErr != null) await taskNotErr; 
            }
            catch (AggregateException ae)
            {
                Console.WriteLine("AggregateException has message: " + ae.Message);
                foreach (var ex in ae.InnerExceptions)
                {
                    Console.WriteLine("I caught exception: " + ex.Message);
                    Console.WriteLine("Inner exception:" + ex.InnerException.Message);
                }
            }
            catch (TaskCanceledException) //if the task start condition is not met, such task is Canceled
            {
                Console.WriteLine("The alternate task has been canceled");
            }
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion


            //            The following extension method “swallows” a task’s unhandled exceptions:

            //            public static void IgnoreExceptions(this Task task)
            //            {
            //                task.ContinueWith(t => { var ignore = t.Exception; },
            //                  TaskContinuationOptions.OnlyOnFaulted);
            //            }
            //(This could be improved by adding code to log the exception.) Here’s how it would be used:

            //Task.Factory.StartNew(() => { throw null; }).IgnoreExceptions();



            #region Continuations and child tasks
            Console.WriteLine("\n--------------------Continuations and child tasks--------");
            TaskCreationOptions atp = TaskCreationOptions.AttachedToParent;
            Task TaskParent = Task.Factory.StartNew(() =>
            {
                Task.Factory.StartNew(() => { throw new Exception("my error 1"); }, atp);
                Task.Factory.StartNew(() => { throw new Exception("my error 2"); }, atp);
                Task.Factory.StartNew(() => { throw new Exception("my error 3"); }, atp);
            })
            .ContinueWith(p => {
                Console.WriteLine("As a parent, I caught something like that: " + p.Exception);
                                throw p.Exception;
            },
                                TaskContinuationOptions.OnlyOnFaulted);
            try
            {
                await TaskParent;
            }
         
            catch (AggregateException ae)
            {
                Console.WriteLine("AggregateException has message: " + ae.Message);
                foreach (var ex in ae.InnerExceptions)
                {
                    Console.WriteLine("I caught exception: " + ex.Message);
                    Console.WriteLine("Inner exception:" + ex.InnerException.Message);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("We have a general exception: ", ex.Message);
            }
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion

            #region Conditional continuations
            Console.WriteLine("\n--------------------Continuation in case of error--------");
            Task t1 = Task.Factory.StartNew(() =>
            {
                Console.WriteLine("t1 works normally but is about to stop");
                //uncoment to simulate error
                //throw new Exception("An error in t1");
            }
            );

            Task fault = t1.ContinueWith(ant => Console.WriteLine("Task run on error t1"),
                                          TaskContinuationOptions.OnlyOnFaulted);

            Task t3 = fault.ContinueWith(ant => Console.WriteLine("t3 worked after the fault")); //the task will run unconditionally whether the fault task executes or not.
            await t3;
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion

            #region A continuation with many predecessors
            Console.WriteLine("\n--------------------A continuation with many predecessors -----");
            Task tA = Task.Factory.StartNew(() => Console.Write("A"));
            Task tB = Task.Factory.StartNew(() => Console.Write("B"));
            Task tContiAB = Task.Factory.ContinueWhenAll(new Task[] { tA, tB }, tasks =>
            { Console.WriteLine("\nContinuation after A and B"); }
            );
            await tContiAB;
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion


            #region Continuation with multiple predecessors and return values
            Console.WriteLine("\n--------------------Continuation with multiple predecessors and return values");
            Task<int> tRetA = Task.Factory.StartNew(() => { Console.WriteLine("A returns 11"); return 11; });
            Task<int> tRetB = Task.Factory.StartNew(() => { Console.WriteLine("B returns 22"); return 22; });
            Task<int> tRetAB = Task.Factory.ContinueWhenAll(new Task<int>[] { tRetA, tRetB }, tasks =>
            { Console.WriteLine($"Continuation after task A wich returns {tasks[0].Result} and  B wich returns {tasks[1].Result}");
                return tasks.Sum(t => t.Result); }
            );
            Console.WriteLine("Sum from tasks A and B is: " + tRetAB.Result);
            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion

            #region continuation with many successors
            Console.WriteLine("\n--------------------continuation with many successors -----");
            Task tPredecessor = Task.Factory.StartNew(() => Console.Write("Predecessor"));
          
            Task tSuccessorA = tPredecessor.ContinueWith(ant =>
            {
                Console.WriteLine("\nContinuation A after predecessor");
            });
            Task tSuccessorB = tPredecessor.ContinueWith(ant =>
            {
                Console.WriteLine("\nContinuation B after predecessor");
            });
            await tSuccessorA;
            await tSuccessorB;

            Console.WriteLine("To continue press ENTER");
            Console.ReadLine();
            #endregion

        }
    }
}
